;; Generated package description from polymode.el  -*- no-byte-compile: t -*-
(define-package "polymode" "0.2.2.0.20251217.132711" "Extensible framework for multiple major modes" '((emacs "25")) :commit "14b1fd8d2a183f11b123f62e02801dc1139da9c1" :maintainer '("Vitalie Spinu" . "spinuvit@gmail.com") :keywords '("languages" "multi-modes" "processes") :url "https://github.com/polymode/polymode")
